"""
Define an API connection object
"""
import logging

import requests
import pendulum

from reposit.data.exceptions import InvalidControllerException
from reposit.data.utils import is_valid_url, deepest_key, match_to_schema
from reposit.settings import BASE_URL

logger = logging.getLogger(__name__)


class ApiRequest(object):
    """
    A class which represents a request/response from the Reposit API.

    Why a class and not a function?
    well we can do some comprehensive validation and checking on creation
    """
    def __str__(self):
        """
        This should give us a good idea (for debugging)
        exactly the request and what data we wanted.
        We don't want to log the controller object as this
        will leak the access token.
        :return:
        """
        return '{} {}'.format(self.url, self.schema)

    def __init__(self, path, controller, schema):
        """
        :param path: the url endpoint (e.g. /v2/deployments etc. etc.
        :param controller: a Controller instance
        :param schema: A dict representing the structure of the response.
        E.g. we want houseP and the API returns the following:
        {
            "data": {
                "houseP": {'blah'}
            }
        }
        So the schema in this case should be a dict like so:
        {
            "data": {
                "houseP": {}
            }
        }
        This is because some of the API responses vary in structure, so
        we can define them on the fly easily :)
        """
        if path.startswith('/'):
            self.url = '{}{}'.format(BASE_URL, path)
        else:
            self.url = '{}/{}'.format(BASE_URL, path)

        assert is_valid_url(self.url)

        if not controller.auth_headers:
            raise InvalidControllerException
        self.controller = controller

        # a lookup of the response schema
        self.schema = schema

    def get(self):
        """
        Once a connection is defined as valid, then a request can be
        made. This formats the response as well as checks the response
        returns an OK http code
        :return:
        """
        resp = requests.get(
            self.url,
            headers=self.controller.auth_headers
        )
        try:
            resp.raise_for_status()
        except Exception as ex:
            # Hijack the exception to log the exact issue.
            logger.exception('Error retrieving data: {}'.format(self))
            raise ex

        data = self._simple_format_for_fields(resp)
        return data

    def query(self, start, end=None):
        """
        Similar to get() but with specified query parameters.
        :param start: unix timestamp (start of query)
        :param end: unix timestamp (end of query). If not specified
        then this defaults to now()
        :return:
        """
        if not end:
            end = pendulum.now().int_timestamp

        resp = requests.get(
            '{}?start={}&end={}'.format(self.url, start, end),
            headers=self.controller.auth_headers
        )
        try:
            resp.raise_for_status()
        except Exception as ex:
            # Hijack the exception to log the exact issue.
            logger.exception('Error retrieving data: {}'.format(self))
            raise ex

        data = self._simple_format_for_fields(resp)
        return {'data': data}

    def _simple_format_for_fields(self, api_response):
        """
        Based on the schema provided, format the data accordingly.

        :return:
        """

        data = api_response.json()
        target_key = deepest_key(self.schema)
        target_data = match_to_schema(data, target_key)
        return target_data



"""
Exceptions relating to auth
"""


class NoAuthenticationError(Exception):
    """
    If a method is called that requires an access token and none is yet
    set then raise this error
    """
    # pylint: disable=unnecessary-pass
    pass



"""
Define the package name and do some shortcut importing so you can just go:

`from reposit import Controller`

"""
from __future__ import absolute_import
import logging
import sys

from reposit.data.controller import Controller
from reposit.auth.account import Account

logger = logging.getLogger(__name__)

"""
Do some basic checks on the python version. We don't wan't anyone using
a deprecated version of Python, (older than 2.7). We also log a warning to
tell a consumer to use Python 3 as it would be good to stop supporting 2
completely. :)
"""
if sys.version_info < (2, 7):
    raise Exception('Versions of Python older than 2.7 are not supported.')

if sys.version_info[0] != 3:
    logger.warning(
        'Python 2 is still supported, but it is recommended you adopt Python 3.'
    )


name = 'reposit'



import pendulum

from reposit.auth import RPConnection
from reposit import Controller, Account

user = RPConnection('repositdemo', 'repositdemo')
account = Account(user)

user_keys = account.get_user_keys()

controller = Controller(user, user_key=user_keys[0])

print(controller.get_remaining_charge(pendulum.now().int_timestamp - 10000))
# print(controller.get_solar_generation_data)


"""
Utility functions for Reposit Client
"""
import six


def dict_iter(_dict):
    """
    Python 3 / Python2 items in a dict
    :param _dict: regular ol' dict
    :return: Representation of the keys/values in a dict in either a 3
    or 2 way.
    """
    return _dict.items() if six.PY3 else _dict.iteritems()



"""
Define an Account for a logged in user.
"""
from reposit.data.api import ApiRequest


class Account(object):
    """
    Represents a user account
    """
    def __init__(self, user):
        self.connection = user
        self.auth_headers = {
            'Authorization': 'Bearer {}'.format(user.token),
            'Reposit-Auth': 'API'
        }

    def get_user_keys(self):
        """
        Retrieve the user key used for subsequent requests
        :param auth_headers:
        :return:
        """
        request = ApiRequest(
            path='/v2/userkeys',
            controller=self,
            schema={'userKeys': {}}
        )
        return request.get()



import pytest

from reposit.data.utils import is_valid_url


@pytest.mark.parametrize('url, expected', [
    ('https://www.repositpower.com', True),
    ('http://www.repositpower.com', True),
    ('://www.repositpower.com', False),
    ('://www.repositpower', False),
    ('www.repositpower', False),
])
def test_is_valid_url(url, expected):
    assert is_valid_url(url) == expected



"""
Defines the class that wraps around the Reposit API
"""
from __future__ import absolute_import

from reposit.data.api import ApiRequest


class Controller(object):
    """
    An object representing a Reposit Box.
    """
    def __init__(self, auth, user_key):

        self.auth_headers = {
            'Authorization': 'Bearer {}'.format(auth.token),
            'Reposit-Auth': 'API'
        }
        self.user_key = user_key
        self.connection = auth

    @property
    def battery_capacity(self):
        """
        Return the kwh battery capacity
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/battery/capacity'.format(self.user_key),
            controller=self,
            schema={'batteryCapacity': {}}
        )
        return request.get()

    @property
    def battery_min_state_of_charge(self):
        """
        Return the minimum state of charge of the battery
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/battery/min_soc'.format(self.user_key),
            controller=self,
            schema={'min_soc': {}}
        )

        return request.get()

    @property
    def has_battery(self):
        """
        If the system has a battery installed
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/components'.format(self.user_key),
            controller=self,
            schema={
                'data': {
                    'battery': {}
                }
            }
        )
        return request.get()

    @property
    def has_inverter(self):
        """
        If the system has an inverter installed
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/components'.format(self.user_key),
            controller=self,
            schema={
                'data': {
                    'inverter': {}
                }
            }
        )
        return request.get()

    def get_solar_generation(self, start, end=None):
        """
        Given a start and end timestamp return the generation
        data.
        :param start: unix timestamp
        :param end: unix timestamp
        :return: list of lists of data
        """
        request = ApiRequest(
            path='v2/deployments/{}/generation/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'solarP': {}
            }
        )
        return request.query(start, end)

    @property
    def latest_solar_generation(self):
        """
        Return a list of data points as lists. Time are in GMT
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/generation/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'solarP': {}
            }
        )
        return request.get()

    def get_house_consumption(self, start, end=None):
        """
        Given a start and end timestamp return the house
        data.
        :param start: unix timestamp
        :param end: unix timestamp
        :return: list of lists of data
        """
        request = ApiRequest(
            path='v2/deployments/{}/house/historical'.format(self.user_key),
            controller=self,
            schema={
                'houseP': {}
            }
        )
        return request.query(start, end)

    @property
    def latest_house_consumption(self):
        """
        Return a list of data points as lists. Time are in GMT
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/house/historical'.format(self.user_key),
            controller=self,
            schema={
                'data': {
                    'houseP': {}
                }
            }
        )
        return request.get()

    def get_battery_data(self, start, end=None):
        """
        Given a start and end timestamp return the inverter
        data.
        :param start: unix timestamp
        :param end: unix timestamp
        :return: list of lists of data
        """
        request = ApiRequest(
            path='v2/deployments/{}/inverter/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'inverterP': {}
            }
        )
        return request.query(start, end)

    @property
    def latest_battery_data(self):
        """
        Return a list of data points as lists. Times are in GMT
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/inverter/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'inverterP': {}
            }
        )
        return request.get()

    def get_remaining_charge(self, start, end=None):
        """
        Given a start and end timestamp return the remaining battery charge
        data.
        :param start: unix timestamp
        :param end: unix timestamp
        :return: list of lists of data
        """
        request = ApiRequest(
            path='v2/deployments/{}/battery/historical/soc'.format(self.user_key),
            controller=self,
            schema={
                'batterySOC': {}
            }
        )
        return request.query(start, end)

    def get_meter_data(self, start, end=None):
        """
        Given a start and end timestamp return the meter
        data.
        :param start: unix timestamp
        :param end: unix timestamp
        :return: list of lists of data
        """
        request = ApiRequest(
            path='v2/deployments/{}/meter/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'meterP': {}
            }
        )
        return request.query(start, end)

    @property
    def latest_meter_data(self):
        """
        Return a list of data points as lists. Times are in GMT
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/meter/historical/p'.format(self.user_key),
            controller=self,
            schema={
                'meterP': {}
            }
        )
        return request.get()

    @property
    def feed_in_tariff(self):
        """
        The feed in tariff cost as a float
        :return:
        """
        request = ApiRequest(
            path='v2/deployments/{}/tariff/tou'.format(self.user_key),
            controller=self,
            schema={
                'fit': {}
            }
        )
        return request.get()



"""
Custom exceptions relating to the usage of the controller class
"""


class InvalidControllerException(Exception):
    """
    A connection is attempted without auth headers
    """
    # pylint: disable=unnecessary-pass
    pass



"""
Establish a connection with the auth endpoints
"""
from __future__ import absolute_import

import logging

import requests
from requests import HTTPError

from requests.auth import HTTPBasicAuth
from reposit.settings import AUTH_PATH

logger = logging.getLogger(__name__)


class RPConnection(object):
    """
    Establish a connection to the Reposit cloud
    """
    def _login(self):
        """
        Given a username and password, obtain an access token
        :return:
        """
        resp = requests.post(AUTH_PATH, auth=HTTPBasicAuth(self.username, self.password), headers={
            "Reposit-Auth": "API"
        })
        try:
            resp.raise_for_status()
        except HTTPError:
            if resp.status_code in (401, 403):
                logger.exception('Unauthorized. Please check your credentials are correct.')
            logger.exception('Unable to authenticate - please try again later.')
            return None, None
        return resp.json()['access_token'], resp.json()['expires_at']

    def __init__(self, username, password):
        """
        Authenticate with a username and password.
        :param username:
        :param password:
        """
        self.username = username
        self.password = password
        self.token, self.token_expiry = self._login()

    def __str__(self):
        return self.username



"""
Utility functions for the reposit controller
"""
from __future__ import absolute_import

import re

# adapted from Django :)
from reposit.utilities import dict_iter

VALID_URL_REGEX = re.compile(
    r'^(?:http|ftp)s?://'  # http:// or https://
)


def is_valid_url(url):
    """
    Check if a url is valid for an api request.
    :param url:
    :return:
    """
    return bool(re.match(VALID_URL_REGEX, url))


def deepest_key(_dict):
    """Return the deepest key in a dict"""
    for key, value in dict_iter(_dict):
        if isinstance(value, dict) and bool(value):
            return deepest_key(_dict[key])
        return key


def match_to_schema(_dict, requested_key):
    """
    Match the schema supplied with the response to return the
    data we requested.
    :param _dict:
    :param requested_key:
    :return:
    """
    if _dict.get(requested_key) is not None:
        return _dict[requested_key]

    for valid_key in _dict:
        if valid_key == requested_key:
            if not isinstance(_dict.get(valid_key), dict):
                return _dict[valid_key]
            else:
                continue
        elif valid_key != requested_key and isinstance(_dict.get(valid_key), dict):
            return match_to_schema(_dict[valid_key], requested_key)
    return None



import pytest

from reposit.data.utils import deepest_key

three_levels = {
    'level1': {
        'level2': {
            'level3': {}
        }
    }
}

single_level = {
    'level1': {}
}

two_levels = {
    'level1': {
        'level2': {}
    }
}


@pytest.mark.parametrize('_dict, expected', [
    (single_level, 'level1'),
    (two_levels, 'level2'),
    (three_levels, 'level3'),
])
def test_deepest_key_func(_dict, expected):
    assert deepest_key(_dict) == expected



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="reposit",
    version="0.4.0",
    author="Thomas Basche",
    author_email="thomas.basche@repositpower.com",
    description="A library to communicate with a Reposit Controller",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RepositPower/reposit-python-client",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests',
        'pendulum',
        'six'
    ],
    classifiers=(
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
)



import pytest

from reposit.data.utils import match_to_schema

key_1 = 'inverterP'

data_1 = {
    'data': {
        'inverterP': 40.0
    }
}

key_2 = 'houseP'

data_2 = {
    'houseP': 50.00
}

key_3 = 'blah2'

data_3 = {
    'data': {
        'generation': {
            'some_data': {
                'blah': 2,
                'blah2': 1,
                'blah3': 40,
            }
        }
    }
}


@pytest.mark.parametrize('_dict, requested_key, expected', [
    (data_1, key_1, 40.0),
    (data_2, key_2, 50.0),
    (data_3, key_3, 1),
])
def test_match_schema_func(_dict, requested_key, expected):
    assert match_to_schema(_dict, requested_key) == expected



summary_response = {
    'data': {
        'battery': {
            'state': 'ONLINE',
            'val': 2.0
        },
        'grid': {
            'state': 'ONLINE',
            'val': 40.0,
            'something_else': 'not_included'
        },
        'house': {
            'state': 'ONLINE',
            'val': 25.0
        },
        'solar': {
            'state': 'ONLINE',
            'val': 30.0,
            'other_key': 'blah'
        },
        'key5': {
            'this_isnt': 'included'
        }
    }
}

expected = {
    'battery': {
        'state': 'ONLINE',
        'value': 2.0
    },
    'grid': {
        'state': 'ONLINE',
        'value': 40.0
    },
    'house': {
        'state': 'ONLINE',
        'value': 25.0
    },
    'solar': {
        'state': 'ONLINE',
        'value': 30.0
    },
}

summary_response_2 = {
    'data': {
        'battery': {
            'state': 'ONLINE',
            'val': 2.0
        },
        'grid': {
            'state': 'ONLINE',
            'val': 40.0,
            'something_else': 'not_included'
        }
    }
}

expected_2 = {
    'battery': {
        'state': 'ONLINE',
        'value': 2.0
    },
    'grid': {
        'state': 'ONLINE',
        'value': 40.0
    }
}